#!/usr/bin/perl

=begin comment
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
=cut

use strict;
use warnings;
use 5.10.1;

use File::Copy;
use File::Temp;
use FindBin;
use Getopt::Std;
use IO::File;
use POSIX qw(strftime);
use Sys::Hostname;
use XML::Writer;

sub HELP_MESSAGE
{
	print STDOUT << "EOF";
Utility to format audit records from ausearch or aureport into a variety of different formats.
Dsitributed under the terms of the GPLv3.
Copyright Nick Sydenham <nsydenham\@yahoo.co.uk>

usage: $0 (-fhilo)
	-f output format, default is XML
	-h display this help
	-i <file> read from file or use stdin if not specified
	-l list available formats
	-o <file> write output to file instead of stdout
	
example: $0 -i audit.log -f XML -o audit.xml
example: ausearch -i -ts recent | $0 -f csv

EOF

	exit;
}

sub listFormatDir
{
	my $dir = shift;
	
	my $tdir = "$FindBin::RealBin/transform/$dir";
	opendir(my $dh, $tdir) || die "Can't open directory $tdir: $!";
	my @files = grep { /\.xslt?$/ && -f "$tdir/$_" } readdir($dh);
    closedir $dh;
	
	my @formats;
	foreach (@files)
	{
		s/[a-zA-Z0-9]+-([a-zA-Z0-9]+)\.xslt?$/$1/i;
		push(@formats, $_);
	}
	push(@formats, "xml (default)");
	
	my @sorted = sort(@formats);
	foreach (@sorted) {
		say " $_";
	}
}

# look in the transform directory for output formats
sub listFormats
{
	say "ausearch formats";
	say "----------------";
	listFormatDir("ausearch");
	
	say "\naureport formats";
	say "----------------";
	listFormatDir("aureport");
	
	exit();
}

use vars qw/ %opt /;
getopts("hli:f:o:", \%opt) or HELP_MESSAGE();
HELP_MESSAGE() if $opt{h};
listFormats() if $opt{l};

# input checking and setup
my $file;
if (defined($opt{i}) && $opt{i} ne "-")
{
	$file = $opt{i};
	die "$file doesn't exist or is empty\n" unless -s $file;
	die "$file is not an audit log file\n" if $file =~ /\.x.lt?$/i;
}

my $input;
if (defined($file)) {
	$input = new IO::File("<$file");
} else {
	$input = \*STDIN;
}

# namespaces
my $ausns = "urn:linux:audit:ausearch";
my $aurns = "urn:linux:audit:aureport";
my $xsi = "http://www.w3.org/2001/XMLSchema-instance";

# XML writer
my $writer;

# write to a temporary file
my $tempfh = File::Temp->new();
my $tempfilename = $tempfh->filename();
#open(my $outfile, ">", $tempfh) or die "Cannot open temporary file: $!";

sub xmlInit
{
	my ($ns) = @_;
	
	# set up the XML Writer
	my %xmlargs = (
		DATA_INDENT => 4,
		DATA_MODE 	=> 1,
		NAMESPACES 	=> 1,
		OUTPUT		=> $tempfh,
		PREFIX_MAP 	=> {$ns => '', $xsi => 'xsi'}
	);
	
	$writer = new XML::Writer(%xmlargs);
	$writer->xmlDecl("UTF-8");
}

# ausearch - parse the type
sub parseType
{
	my $prefix = shift;
	
	my ($type, $tstamp) = $prefix =~ /type=([A-Z_]+) msg=.*?\((.*)/;
															  
	if ($prefix =~ /^node/) {
		my ($node) = $prefix =~ /node=([a-zA-Z0-9._-]+)/;
		$writer->startTag([$ausns, "record"], "node" => $node, "type" => $type, "timestamp" => $tstamp);
	} else {
		$writer->startTag([$ausns, "record"], "type" => $type, "timestamp" => $tstamp);
	}
}

# ausearch - parse the detail
sub parseDetail
{
	my $detail = shift;
	my $attr;
	
	my @elements = split(/ /, $detail);
	while ($#elements >= 0)
	{
		my $el = shift @elements;

		if (length($el) == 0) {
			next;
		}
		elsif ($el !~ /=/) # info element - maybe multiple entries
		{
			my $info = $el;
			
			if ($info eq "old" || $info eq "new")
			{
				$attr = $info;
				next;
			}
				
			my $nextinfo;
			if ($#elements > 0)
			{
				$nextinfo = shift(@elements);
				while ($nextinfo !~ /=/)
				{
					$info = "$info $nextinfo";
					$nextinfo = shift(@elements);
				}
			}
			
			if (defined($nextinfo) && $nextinfo =~ /=/) {
				unshift(@elements, $nextinfo);
			}
			
			$writer->dataElement([$ausns, "info"], $info);
		}
		else
		{
			my ($arg, $val) = split(/=/, $el, 2);
			
			# check for a value
			if ($val =~ /^["(].*[^")]$/) {
				do {
					$val = "$val ".shift(@elements);
				} while ($val !~ /[")]$/);
			}
			
			
			if (defined($attr)) {
				$writer->dataElement([$ausns, $arg], $val, "status" => $attr);
				undef($attr);
			} else {
				$writer->dataElement([$ausns, $arg], $val);
			}
		}
	}	
}

my $now = time();
# We need to munge the timezone indicator to add a colon between the hour and minute part
my $tz = strftime("%z", localtime($now));
$tz =~ s/(\d{2})(\d{2})/$1:$2/;
my $time = strftime("%Y-%m-%dT%H:%M:%S", localtime($now)) . $tz;

my $line = 0;
my $record = 0; # ausearch - used to indicate processing a record
my @header; # aureport - header fields

my $AUSEARCH = 1;
my $AUREPORT = 2;
my $autype = 0;

while (<$input>)
{
	$line++;

	# figure out what is being processed
	if ($line == 1)
	{
		if (/----/)
		{
			$autype = $AUSEARCH;
			xmlInit($ausns);
			$writer->startTag([$ausns, "ausearch"],
							  "hostname" => hostname(), "generated" => $time, "generator" => "auformat",
							  [$xsi, "schemaLocation"] => "$ausns audits.xsd"
			);
		}
		else
		{
			$autype = $AUREPORT;
			xmlInit($aurns);
			next;
		}
	}

	# ausearch
	if ($autype == $AUSEARCH)
	{
		if (/----/)
		{
			if ($record)
			{
				$writer->endTag();
				$record = 0;
			}
			$writer->startTag([$ausns, "Event"]);
			$record = 1;
		}
		elsif (/^[a-z]+=/)
		{
			my ($type, $detail) = /^(.*?)\)\s+:\s+(.*)$/;
			
			parseType($type);
			
			# add a space before opening brackets
			$detail =~ s/([^ =])\(/$1 \(/g;
			
			# remove double spaces
			$detail =~ s/\s\s/ /g;
			
			if ($detail =~ /msg=\'/)
			{
				my ($before, $msg) = ($detail =~ /(.*)msg=\'(.*)'/);
				
				parseDetail($before);
				$writer->startTag([$ausns, "msg"]);
				parseDetail($msg);
				$writer->endTag();
			}
			else
			{
				parseDetail($detail);
			}
			
			$writer->endTag();
		}
		else {
			say STDERR "Invalid line found at line $line";
		}
	}
	# aureport
	else
	{
		if (/^[0-9]/)
		{
			my @fields = split;
			chop($fields[0]); # remove .
			$writer->startTag([$aurns, "record"], "num" => $fields[0]);
			
			my $fnum=1;
			for (@fields[1..$#fields]) {
				$writer->dataElement([$aurns, "field"], $_, "col" => $header[$fnum++]);
			}
			$writer->endTag();
		}
		elsif (/^#/)
		{
			@header = split;
			$writer->startTag([$aurns, "header"]);
			for (@header[1..$#header])	{
				$writer->dataElement([$aurns, "col"], $_);
			}
			$writer->endTag();
		}
		elsif (/^[A-Z]/)
		{
			chomp($_);
			$writer->startTag([$aurns, "aureport"],
							"hostname" => hostname(), "generated" => $time,
							"generator" => "auformat", "title" => $_,
							[$xsi, "schemaLocation"] => "$aurns auditr.xsd"
			);
		}
	}
}
if ($autype == $AUSEARCH) {
	$writer->endTag();
}
$writer->endTag();
$writer->end();

close($input);
$tempfh->close();

# check for formatting
my $format = "xml";
my $tdir = ($autype == $AUSEARCH ? "ausearch" : "aureport");
if ($opt{f} && lc($opt{f}) ne "xml")
{
	$format = lc($opt{f});
	die "Format $format doesn't exist" unless (-f "$FindBin::RealBin/transform/$tdir/$format.xsl" || $format eq "xml");
}

# formatting
if ($format eq "xml")
{
	if ($opt{o}) {
		copy($tempfilename, $opt{o});
	} else {
		open(my $fh, "<", $tempfilename) or die "Cannot open temp file: $!";
		while (<$fh>) {
			print;
		}
		close($fh);
	}
	exit();
}

# transform the XML
my $transform = "$FindBin::RealBin/transform/$tdir/$format.xsl";

# try the Perl module XML::LibXSLT first
eval
{
	require XML::LibXSLT;
	my $source = XML::LibXML->load_xml(location => $tempfilename);
	my $style = XML::LibXML->load_xml(location => $transform, no_cdata=>1);
	my $xslt = XML::LibXSLT->new();
	my $stylesheet = $xslt->parse_stylesheet($style);
	my $results = $stylesheet->transform($source);
	if ($opt{o}) {
		$stylesheet->output_file($results, $opt{o});
	} else {
		print $stylesheet->output_as_bytes($results);
	}
};

# try xsltproc from libxslt-tools
if ($@)
{
	if ($opt{o}) {
		system("xsltproc", "-o", $opt{o}, $transform, $tempfilename) or die "Cannot transform. Please install either Perl's XML::LibXML or libxslt-tools";
	} else {
		system("xsltproc", $transform, $tempfilename) or die "Cannot transform. Please install either Perl's XML::LibXML or libxslt-tools";
	}
}
unlink $tempfilename;
